# 1.2.2树递归

> 译者： [https://sicp.comp.nus.edu.sg/chapters/13](https://sicp.comp.nus.edu.sg/chapters/13)

另一种常见的计算模式称为_树递归_。 例如，考虑计算斐波纳契数的顺序，其中每个数是前两个数字的和：![%200%2C%201%2C%201%2C%202%2C%203%2C%205%2C%208%2C%2013%2C%2021%2C%20%5Cldots%20](img/54b981a703b188ca100cff785309bfa7.jpg)通常，斐波那契数可以由规则\ begin {eqnarray *} \ textrm {Fib}定义 （n）& = & \ left \ {\ begin {array} {ll} 0 & \ mbox {if ![n%3D0](img/f55ff3a5c44f073e60c4ff4c95c776c5.jpg)} \\ 1 & \ mbox {if ![n%3D1](img/c74b8369318b12ee6e04beba86ba15f3.jpg)} \\ \ textrm {Fib}（n-1）+ \ textrm {Fib}（n-2）& \ mbox {否则} \ end {array} \ right。 \ end {eqnarray *}

我们可以立即将此定义转换为用于计算斐波那契数的递归函数：

```js
function fib(n) {
    return n === 0
           ? 0
           : n === 1
             ? 1
             : fib(n - 1) + fib(n - 2);
}
```

<figure><split>![](img/1b1d1252fd2598d45849e41da9cf8353.jpg)</split> **[Figure 1.5](13#fig_1.5)** The tree-recursive process generated in computing `fib(5)` .</figure>

考虑这种计算的模式。 为了计算`fib(5)`，我们计算`fib(4)`和`fib(3)`。 为了计算`fib(4)`，我们计算`fib(3)`和`fib(2)`。 通常，进化的过程看起来像一棵树，如图 [1.5](13#fig_1.5) 所示。 请注意，分支在每个级别均分为两部分（底部除外）； 这反映了一个事实，即`fib`函数每次被调用两次。

此函数对于原型树的递归来说是有启发性的，但是它是计算斐波纳契数的一种糟糕方法，因为它执行了很多冗余计算。 在图 [1.5](13#fig_1.5)[ 1.13 ](13#ex_1.13) ），![%5Ctextrm%7BFib%7D%28n%29](img/23b0184178f6f889c4726a0650f99104.jpg)是最接近![%5Cphi%5E%7Bn%7D%20%2F%5Csqrt%7B5%7D](img/95fe4b33e0e38ee6ad921ca2a1a98c68.jpg)的整数，其中![%20%5Cphi%3D%281%2B%5Csqrt%7B5%7D%29%2F2%5Capprox%201.6180%20](img/2f7dc1fdd21498744e3b5dc563d257d7.jpg)是_黄金分割率_，满足 公式![%20%5Cphi%5E%7B2%7D%20%3D%5Cphi%20%2B%201%20](img/aa7cfac5b4dc8e608673c344bd190a01.jpg)因此，该过程使用许多步骤，这些步骤与输入呈指数增长。 另一方面，所需空间仅随输入线性增长，因为在计算的任何时候，我们只需要跟踪树中哪些节点在我们之上即可。 通常，树递归过程所需的步骤数将与树中的节点数成正比，而所需的空间将与树的最大深度成正比。

我们还可以制定一个迭代过程来计算斐波那契数。 这个想法是使用一对初始化为![%5Ctextrm%7BFib%7D%281%29%3D1](img/f9f0fcc7f0fd1a962d9de25c03997627.jpg)和![%5Ctextrm%7BFib%7D%280%29%3D0](img/052163656c9d1a76dd899efa481efce4.jpg)的整数![a](img/070b1af5eca3a5c5d72884b536090f17.jpg)和![b](img/6872867a863714d15d9a0d64c20734ce.jpg)，并重复应用同时转换\ begin {eqnarray *} a & \ leftarrow & a + b \\ b & \ leftarrow & a \ end {eqnarray *}不难证明，在应用此变换![n](img/493731e423d5db62086d0b8705dda0c8.jpg)次后，![a](img/070b1af5eca3a5c5d72884b536090f17.jpg)和![b](img/6872867a863714d15d9a0d64c20734ce.jpg)分别相等， 到![%5Ctextrm%7BFib%7D%28n%2B1%29](img/ca23e2b311a6ee02fe97321bd59e2eef.jpg)和![%5Ctextrm%7BFib%7D%28n%29](img/23b0184178f6f889c4726a0650f99104.jpg)。 因此，我们可以使用函数迭代计算斐波那契数

```js
function fib(n) {
    return fib_iter(1, 0, n);
}
function fib_iter(a, b, count) {
    return count === 0
           ? b
           : fib_iter(a + b, a, count - 1);
}
```

This second method for computing ![%5Ctextrm%7BFib%7D%28n%29](img/23b0184178f6f889c4726a0650f99104.jpg) is a linear iteration. The difference in number of steps required by the two methods—one linear in ![n](img/493731e423d5db62086d0b8705dda0c8.jpg), one growing as fast as ![%5Ctextrm%7BFib%7D%28n%29](img/23b0184178f6f889c4726a0650f99104.jpg) itself—is enormous, even for small inputs.

不能由此得出结论，树递归过程是无用的。 当我们考虑对分层结构的数据而不是数字进行操作的流程时，我们会发现树递归是一种自然而强大的工具。 [[1]](13#footnote-1) 但是，即使在数值运算中，树递归过程也可以帮助我们理解和设计程序。 例如，尽管第一个`fib`函数的效率比第二个函数低得多，但它更直接，只不过是将Fibonacci序列的定义翻译成JavaScript而已。 要制定迭代算法，需要注意，可以将计算重铸为具有三个状态名称的迭代。

## 示例：计算变更

提出迭代的斐波那契算法只需要一点技巧。 相反，请考虑以下问题：![%5C](img/c9fb3ce835498ee95f920d1caf849660.jpg) 1.00有多少种不同的更改方法？![%2C%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20given%20half-dollars%2C%20quarters%2C%20dimes%2C%20nickels%2C%20and%20pennies%3F%20%20More%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20generally%2C%20can%20we%20write%20a%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20function%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20to%20compute%20the%20number%20of%20ways%20to%20change%20any%20given%20amount%20of%20money%3F%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22p8%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cp%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20This%20problem%20has%20a%20simple%20solution%20as%20a%20recursive%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20function.%20%20Suppose%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20we%20think%20of%20the%20types%20of%20coins%20available%20as%20arranged%20in%20some%20order.%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20Then%20the%20following%20relation%20holds%3A%0A%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20The%20number%20of%20ways%20to%20change%20amount%20%3Cimg%20alt%3D%22a%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Da%22%20%2F%3E%20using%20%3Cimg%20alt%3D%22n%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Dn%22%20%2F%3E%20kinds%20of%20coins%20equals%0A%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Cul%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Cli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20the%20number%20of%20ways%20to%20change%20amount%20%3Cimg%20alt%3D%22a%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Da%22%20%2F%3E%20using%20all%20but%20the%20first%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20kind%20of%20coin%2C%20plus%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Fli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Cli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20the%20number%20of%20ways%20to%20change%20amount%20%3Cimg%20alt%3D%22a-d%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Da-d%22%20%2F%3E%20using%20all%20%3Cimg%20alt%3D%22n%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Dn%22%20%2F%3E%20kinds%20of%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20coins%2C%20where%20%3Cimg%20alt%3D%22d%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Dd%22%20%2F%3E%20is%20the%20denomination%20of%20the%20first%20kind%20of%20coin.%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Fli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Ful%3E%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22p9%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cp%3E%0A%20%20%20%20%20%20%20%20%20%20To%20see%20why%20this%20is%20true%2C%20observe%20that%20the%20ways%20to%20make%20change%20can%20be%0A%20%20%20%20%20%20%20%20%20%20divided%20into%20two%20groups%3A%20those%20that%20do%20not%20use%20any%20of%20the%20first%20kind%0A%20%20%20%20%20%20%20%20%20%20of%20coin%2C%20and%20those%20that%20do.%20%20Therefore%2C%20the%20total%20number%20of%20ways%20to%0A%20%20%20%20%20%20%20%20%20%20make%20change%20for%20some%20amount%20is%20equal%20to%20the%20number%20of%20ways%20to%20make%0A%20%20%20%20%20%20%20%20%20%20change%20for%20the%20amount%20without%20using%20any%20of%20the%20first%20kind%20of%20coin%2C%0A%20%20%20%20%20%20%20%20%20%20plus%20the%20number%20of%20ways%20to%20make%20change%20assuming%20that%20we%20do%20use%20the%0A%20%20%20%20%20%20%20%20%20%20first%20kind%20of%20coin.%20%20But%20the%20latter%20number%20is%20equal%20to%20the%20number%20of%0A%20%20%20%20%20%20%20%20%20%20ways%20to%20make%20change%20for%20the%20amount%20that%20remains%20after%20using%20a%20coin%20of%0A%20%20%20%20%20%20%20%20%20%20the%20first%20kind.%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22p10%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cp%3E%0A%20%20%20%20%20%20%20%20%20%20Thus%2C%20we%20can%20recursively%20reduce%20the%20problem%20of%20changing%20a%20given%20amount%0A%20%20%20%20%20%20%20%20%20%20to%20problems%20of%20changing%20smaller%20amounts%20or%20using%20fewer%20kinds%20of%20coins.%0A%20%20%20%20%20%20%20%20%20%20Consider%20this%20reduction%20rule%20carefully%2C%20and%20convince%20yourself%20that%20we%0A%20%20%20%20%20%20%20%20%20%20can%20use%20it%20to%20describe%20an%20algorithm%20if%20we%20specify%20the%20following%0A%20%20%20%20%20%20%20%20%20%20degenerate%20cases%3A%3Ca%20class%3D%22superscript%22%20id%3D%22footnote-link-2%22%20href%3D%2213%23footnote-2%22%3E%5B2%5D%3C%2Fa%3E%0A%0A%20%20%20%20%20%20%20%20%20%20%3Cul%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%3Cli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20If%20%3Cimg%20alt%3D%22a%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Da%22%20%2F%3E%20is%20exactly%200%2C%20we%20should%20count%20that%20as%201%20way%20to%20make%20change.%0A%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Fli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%3Cli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20If%20%3Cimg%20alt%3D%22a%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Da%22%20%2F%3E%20is%20less%20than%200%2C%20we%20should%20count%20that%20as%200%20ways%20to%20make%20change.%0A%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Fli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%3Cli%3E%20If%20%3Cimg%20alt%3D%22n%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Dn%22%20%2F%3E%20is%200%2C%20we%20should%20count%20that%20as%200%20ways%20to%20make%20change.%0A%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Fli%3E%0A%20%20%20%20%20%20%20%20%20%20%3C%2Ful%3E%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22p11%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cp%3E%0A%20%20%20%20%20%20%20%20%20%20We%20can%20easily%20translate%20this%20description%20into%20a%20recursive%0A%20%20%20%20%20%20%20%20%20%20function%3A%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22snippet%22%20id%3D%22javascript_13_2_div%22%3E%3Cdiv%20class%3D%22pre-prettyprint%22%3E%3Cpre%20class%3D%22prettyprint%22%20title%3D%22Evaluate%20Javascript%20expression%22%20onclick%3D%22var%20compressed%20%3D%20LZString.compressToEncodedURIComponent%28''%2B'%5Cn'%2B'function%20count_change%28amount%29%20%7B%5Cn%20%20%20%20return%20cc%28amount%2C%205%29%3B%5Cn%7D%5Cnfunction%20cc%28amount%2C%20kinds_of_coins%29%20%7B%5Cn%20%20%20%20return%20amount%20%3D%3D%3D%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%3F%201%5Cn%20%20%20%20%20%20%20%20%20%20%20%3A%20amount%20%3C%200%20%7C%7C%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%3F%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%3A%20cc%28amount%2C%20kinds_of_coins%20-%201%29%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%2B%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20cc%28amount%20-%20first_denomination%28%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%29%2C%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%29%3B%5Cn%7D%5Cnfunction%20first_denomination%28kinds_of_coins%29%20%7B%5Cn%20%20%20%20return%20kinds_of_coins%20%3D%3D%3D%201%20%3F%201%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%202%20%3F%205%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%203%20%3F%2010%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%204%20%3F%2025%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%205%20%3F%2050%20%3A%200%3B%5Cn%7D'%2B'%5Cn'%2B'%5Cncount_change%28100%29%3B%5Cn%20%20%20%20%20%20%20%20%20%20'%2B'%5Cn'%29%3B%20var%20url%20%3D%20'https%3A%2F%2Fsource-academy.github.io%2Fplayground%23chap%3D1%26prgrm%3D'%2Bcompressed%3B%20window.open%28url%29%3B%20%22%3Efunction%20count_change%28amount%29%20%7B%0A%20%20%20%20return%20cc%28amount%2C%205%29%3B%0A%7D%0Afunction%20cc%28amount%2C%20kinds_of_coins%29%20%7B%0A%20%20%20%20return%20amount%20%3D%3D%3D%200%0A%20%20%20%20%20%20%20%20%20%20%20%3F%201%0A%20%20%20%20%20%20%20%20%20%20%20%3A%20amount%20%3C%200%20%7C%7C%0A%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%200%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%3F%200%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%3A%20cc%28amount%2C%20kinds_of_coins%20-%201%29%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%2B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20cc%28amount%20-%20first_denomination%28%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%29%2C%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%29%3B%0A%7D%0Afunction%20first_denomination%28kinds_of_coins%29%20%7B%0A%20%20%20%20return%20kinds_of_coins%20%3D%3D%3D%201%20%3F%201%20%3A%0A%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%202%20%3F%205%20%3A%0A%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%203%20%3F%2010%20%3A%0A%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%204%20%3F%2025%20%3A%0A%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%205%20%3F%2050%20%3A%200%3B%0A%7D%3C%2Fpre%3E%3C%2Fdiv%3E%3C%2Fdiv%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22p12%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cp%3E%0A%20%20%20%20%20%20%20%20%20%20%28The%0A%20%20%20%20%20%20%20%20%20%20%3Ccode%3Efirst_denomination%3C%2Fcode%3E%0A%20%20%20%20%20%20%20%20%20%20function%0A%20%20%20%20%20%20%20%20%20%20takes%20as%20input%20the%20number%20of%0A%20%20%20%20%20%20%20%20%20%20kinds%20of%20coins%20available%20and%20returns%20the%20denomination%20of%20the%20first%0A%20%20%20%20%20%20%20%20%20%20kind.%20%20Here%20we%20are%20thinking%20of%20the%20coins%20as%20arranged%20in%20order%20from%0A%20%20%20%20%20%20%20%20%20%20largest%20to%20smallest%2C%20but%20any%20order%20would%20do%20as%20well.%29%20%20We%20can%20now%0A%20%20%20%20%20%20%20%20%20%20answer%20our%20original%20question%20about%20changing%20a%20dollar%3A%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22snippet%22%20id%3D%22javascript_13_3_div%22%3E%3Cdiv%20class%3D%22pre-prettyprint%22%3E%3Cpre%20class%3D%22prettyprint%22%20title%3D%22Evaluate%20Javascript%20expression%22%20onclick%3D%22var%20compressed%20%3D%20LZString.compressToEncodedURIComponent%28'%5Cnfunction%20count_change%28amount%29%20%7B%5Cn%20%20%20%20return%20cc%28amount%2C%205%29%3B%5Cn%7D%5Cnfunction%20cc%28amount%2C%20kinds_of_coins%29%20%7B%5Cn%20%20%20%20return%20amount%20%3D%3D%3D%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%3F%201%5Cn%20%20%20%20%20%20%20%20%20%20%20%3A%20amount%20%3C%200%20%7C%7C%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%3F%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%3A%20cc%28amount%2C%20kinds_of_coins%20-%201%29%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%2B%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20cc%28amount%20-%20first_denomination%28%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%29%2C%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%29%3B%5Cn%7D%5Cnfunction%20first_denomination%28kinds_of_coins%29%20%7B%5Cn%20%20%20%20return%20kinds_of_coins%20%3D%3D%3D%201%20%3F%201%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%202%20%3F%205%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%203%20%3F%2010%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%204%20%3F%2025%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%205%20%3F%2050%20%3A%200%3B%5Cn%7D%5Cn%20%20%20%20%20%20%20%20%20%20'%2B'%5Cn'%2B'count_change%28100%29%3B'%2B'%5Cn'%2B''%2B'%5Cn'%29%3B%20var%20url%20%3D%20'https%3A%2F%2Fsource-academy.github.io%2Fplayground%23chap%3D1%26prgrm%3D'%2Bcompressed%3B%20window.open%28url%29%3B%20%22%3Ecount_change%28100%29%3B%3C%2Fpre%3E%3C%2Fdiv%3E%3C%2Fdiv%3E%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22p13%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cp%3E%0A%20%20%20%20%20%20%20%20%20%20The%20function%20%3Ccode%3Ecount_change%3C%2Fcode%3E%0A%20%20%20%20%20%20%20%20%20%20generates%20a%20tree-recursive%20process%20with%0A%20%20%20%20%20%20%20%20%20%20redundancies%20similar%20to%20those%20in%20our%20first%20implementation%20of%0A%20%20%20%20%20%20%20%20%20%20%3Ccode%3Efib%3C%2Fcode%3E.%0A%20%20%20%20%20%20%20%20%20%20%28It%20will%20take%20quite%20a%20while%20for%20that%20292%20to%20be%20computed.%29%20%20On%0A%20%20%20%20%20%20%20%20%20%20the%20other%20hand%2C%20it%20is%20not%20obvious%20how%20to%20design%20a%20better%20algorithm%0A%20%20%20%20%20%20%20%20%20%20for%20computing%20the%20result%2C%20and%20we%20leave%20this%20problem%20as%20a%20challenge.%0A%20%20%20%20%20%20%20%20%20%20The%20observation%20that%20a%0A%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20tree-recursive%20process%20may%20be%20highly%0A%20%20%20%20%20%20%20%20%20%20inefficient%20but%20often%20easy%20to%20specify%20and%20understand%20has%20led%20people%20to%0A%20%20%20%20%20%20%20%20%20%20propose%20that%20one%20could%20get%20the%20best%20of%20both%20worlds%20by%20designing%20a%0A%20%20%20%20%20%20%20%20%20%20%3Cquote%3Esmart%20compiler%3C%2Fquote%3E%20that%20could%20transform%20tree-recursive%0A%20%20%20%20%20%20%20%20%20%20functions%0A%20%20%20%20%20%20%20%20%20%20into%20more%20efficient%0A%20%20%20%20%20%20%20%20%20%20functions%0A%20%20%20%20%20%20%20%20%20%20that%20compute%20the%20same%20result.%3Ca%20class%3D%22superscript%22%20id%3D%22footnote-link-3%22%20href%3D%2213%23footnote-3%22%3E%5B3%5D%3C%2Fa%3E%0A%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22ex_1.11%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cexercise%3E%3Cb%3E%3Ca%20class%3D%22exercise-number%20permalink%22%20id%3D%22ex_1.11%22%3EExercise%201.11%20%3C%2Fa%3E%3C%2Fb%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20A%20function%20%3Cimg%20alt%3D%22f%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Df%22%20%2F%3E%20is%20defined%20by%20the%20rule%20that%20%3Cimg%20alt%3D%22f%2528n%2529%253Dn%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Df%2528n%2529%253Dn%22%20%2F%3E%20if%20%3Cimg%20alt%3D%22n%253C3%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Dn%253C3%22%20%2F%3E%20and%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Cimg%20alt%3D%22f%2528n%2529%253D%257Bf%2528n-1%2529%257D%252B2f%2528n-2%2529%252B3f%2528n-3%2529%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Df%2528n%2529%253D%257Bf%2528n-1%2529%257D%252B2f%2528n-2%2529%252B3f%2528n-3%2529%22%20%2F%3E%20if%20%3Cimg%20alt%3D%22n%255Cge%25203%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Dn%255Cge%25203%22%20%2F%3E.%20%20Write%20a%20JavaScript%20function%20that%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20computes%20%3Cimg%20alt%3D%22f%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Df%22%20%2F%3E%20by%20means%20of%20a%20recursive%20process.%20%20Write%20a%20function%20that%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20computes%20%3Cimg%20alt%3D%22f%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Df%22%20%2F%3E%20by%20means%20of%20an%20iterative%20process.%0A%3Cdiv%20class%3D%22Solution%22%3E%0A%3Cdiv%20class%3D%22solution_btn%22%3E%3Cbutton%20class%3D%22btn%20btn-secondary%20solution_btn%22%20href%3D%22%23solution_13_1_div%22%20data-toggle%3D%22collapse%22%3ESolution%3C%2Fbutton%3E%3C%2Fdiv%3E%0A%3Cdiv%20class%3D%22solution_content%20collapse%22%20id%3D%22solution_13_1_div%22%3E%3Csolution%3E%0A%20%20%20%20%20%20%20%3Cdiv%20class%3D%22snippet%22%20id%3D%22javascript_13_4_div%22%3E%3Cdiv%20class%3D%22pre-prettyprint%22%3E%3Cpre%20class%3D%22prettyprint%22%20title%3D%22Evaluate%20Javascript%20expression%22%20onclick%3D%22var%20compressed%20%3D%20LZString.compressToEncodedURIComponent%28''%2B'%5Cn'%2B'%2F%2F%20iterative%20function%5Cnfunction%20f_iterative%28n%29%20%7B%5Cn%20%20%20%20return%20n%20%3C%203%5Cn%20%20%20%20%20%20%20%20%20%20%20%3F%20n%5Cn%20%20%20%20%20%20%20%20%20%20%20%3A%20f_iterative_impl%282%2C%201%2C%200%2C%20n%20-%202%29%3B%5Cn%7D%5Cnfunction%20f_iterative_impl%28a%2C%20b%2C%20c%2C%20count%29%20%7B%5Cn%20%20%20%20return%20count%20%3D%3D%3D%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%3F%20a%5Cn%20%20%20%20%20%20%20%20%20%20%20%3A%20f_iterative_impl%28a%20%2B%202%20*%20b%20%2B%203%20*%20c%2C%20a%2C%20b%2C%20count%20-%201%29%3B%5Cn%7D'%2B'%5Cn'%2B'%5Cnf_iterative%285%29%3B%5Cn%20%20%20%20%20%20%20'%2B'%5Cn'%29%3B%20var%20url%20%3D%20'https%3A%2F%2Fsource-academy.github.io%2Fplayground%23chap%3D1%26prgrm%3D'%2Bcompressed%3B%20window.open%28url%29%3B%20%22%3E%2F%2F%20iterative%20function%0Afunction%20f_iterative%28n%29%20%7B%0A%20%20%20%20return%20n%20%3C%203%0A%20%20%20%20%20%20%20%20%20%20%20%3F%20n%0A%20%20%20%20%20%20%20%20%20%20%20%3A%20f_iterative_impl%282%2C%201%2C%200%2C%20n%20-%202%29%3B%0A%7D%0Afunction%20f_iterative_impl%28a%2C%20b%2C%20c%2C%20count%29%20%7B%0A%20%20%20%20return%20count%20%3D%3D%3D%200%0A%20%20%20%20%20%20%20%20%20%20%20%3F%20a%0A%20%20%20%20%20%20%20%20%20%20%20%3A%20f_iterative_impl%28a%20%2B%202%20*%20b%20%2B%203%20*%20c%2C%20a%2C%20b%2C%20count%20-%201%29%3B%0A%7D%3C%2Fpre%3E%3C%2Fdiv%3E%3C%2Fdiv%3E%0A%0A%3Cdiv%20class%3D%22snippet%22%20id%3D%22javascript_13_5_div%22%3E%3Cdiv%20class%3D%22pre-prettyprint%22%3E%3Cpre%20class%3D%22prettyprint%22%20title%3D%22Evaluate%20Javascript%20expression%22%20onclick%3D%22var%20compressed%20%3D%20LZString.compressToEncodedURIComponent%28''%2B'%5Cn'%2B'%2F%2Frecursive%20function%5Cnfunction%20f_recursive%28n%29%20%7B%5Cn%20%20%20%20return%20n%20%3C%203%5Cn%20%20%20%20%20%20%20%20%20%20%20%3F%20n%5Cn%20%20%20%20%20%20%20%20%20%20%20%3A%20f_recursive%28n%20-%201%29%20%2B%5Cn%09%20%20%20%20%202%20*%20f_recursive%28n%20-%202%29%20%2B%5Cn%09%20%20%20%20%203%20*%20f_recursive%28n%20-%203%29%3B%5Cn%7D'%2B'%5Cn'%2B'%5Cnf_recursive%285%29%3B%5Cn%20%20%20%20%20%20%20'%2B'%5Cn'%29%3B%20var%20url%20%3D%20'https%3A%2F%2Fsource-academy.github.io%2Fplayground%23chap%3D1%26prgrm%3D'%2Bcompressed%3B%20window.open%28url%29%3B%20%22%3E%2F%2Frecursive%20function%0Afunction%20f_recursive%28n%29%20%7B%0A%20%20%20%20return%20n%20%3C%203%0A%20%20%20%20%20%20%20%20%20%20%20%3F%20n%0A%20%20%20%20%20%20%20%20%20%20%20%3A%20f_recursive%28n%20-%201%29%20%2B%0A%09%20%20%20%20%202%20*%20f_recursive%28n%20-%202%29%20%2B%0A%09%20%20%20%20%203%20*%20f_recursive%28n%20-%203%29%3B%0A%7D%3C%2Fpre%3E%3C%2Fdiv%3E%3C%2Fdiv%3E%0A%3C%2Fsolution%3E%3C%2Fdiv%3E%0A%3C%2Fdiv%3E%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%3C%2Fexercise%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22ex_1.12%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cexercise%3E%3Cb%3E%3Ca%20class%3D%22exercise-number%20permalink%22%20id%3D%22ex_1.12%22%3EExercise%201.12%20%3C%2Fa%3E%3C%2Fb%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20The%20following%20pattern%20of%20numbers%20is%20called%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Cem%3EPascal's%20triangle%3C%2Fem%3E.%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Ctreetab%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Cimg%20alt%3D%22%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%257B%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%255Cbegin%257Barray%257D%257Brrrrrrrrr%257D%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%25201%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2526%25202%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%25263%2520%2520%2526%2520%2520%2520%25263%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%25201%2520%2520%2526%2520%2520%2520%2526%25204%2520%2526%2520%2520%2520%2526%25206%2520%2526%2520%2520%2520%2526%25204%2520%2526%2520%2520%2520%2526%25201%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%255Cldots%2520%2526%2520%2526%2520%2526%2520%2526%2520%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%255Cend%257Barray%257D%257D%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3D%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%257B%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%255Cbegin%257Barray%257D%257Brrrrrrrrr%257D%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%25201%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2526%25202%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%25263%2520%2520%2526%2520%2520%2520%25263%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%25201%2520%2520%2526%2520%2520%2520%2526%25204%2520%2526%2520%2520%2520%2526%25206%2520%2526%2520%2520%2520%2526%25204%2520%2526%2520%2520%2520%2526%25201%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%255Cldots%2520%2526%2520%2526%2520%2526%2520%2526%2520%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%255Cend%257Barray%257D%257D%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%22%20%2F%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Ftreetab%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20The%20numbers%20at%20the%20edge%20of%20the%20triangle%20are%20all%201%2C%20and%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20each%20number%20inside%20the%20triangle%20is%20the%20sum%20of%20the%20two%20numbers%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20above%20it.%3Ca%20class%3D%22superscript%22%20id%3D%22footnote-link-4%22%20href%3D%2213%23footnote-4%22%3E%5B4%5D%3C%2Fa%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20Write%20a%20function%20that%20computes%20elements%20of%20Pascal's%20triangle%20by%20means%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20of%20a%20recursive%20process.%0A%3Cdiv%20class%3D%22Solution%22%3E%0A%3Cdiv%20class%3D%22solution_btn%22%3E%3Cbutton%20class%3D%22btn%20btn-secondary%20solution_btn%22%20href%3D%22%23solution_13_2_div%22%20data-toggle%3D%22collapse%22%3ESolution%3C%2Fbutton%3E%3C%2Fdiv%3E%0A%3Cdiv%20class%3D%22solution_content%20collapse%22%20id%3D%22solution_13_2_div%22%3E%3Csolution%3E%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22snippet%22%20id%3D%22javascript_13_6_div%22%3E%3Cdiv%20class%3D%22pre-prettyprint%22%3E%3Cpre%20class%3D%22prettyprint%22%20title%3D%22Evaluate%20Javascript%20expression%22%20onclick%3D%22var%20compressed%20%3D%20LZString.compressToEncodedURIComponent%28''%2B'%5Cn'%2B'function%20pascal_triangle%28row%2C%20index%29%20%7B%5Cn%20%20%20%20return%20index%20%3E%20row%5Cn%20%20%20%20%20%20%20%20%20%20%20%3F%20false%5Cn%20%20%20%20%20%20%20%20%20%20%20%3A%20index%20%3D%3D%3D%201%20%7C%7C%20index%3D%3D%3Drow%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%3F%201%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%3A%20pascal_triangle%28row%20-%201%2C%20index%20-%201%29%5Cn%09%20%20%20%20%20%20%20%2B%5Cn%09%20%20%20%20%20%20%20pascal_triangle%28row%20-%201%2C%20index%29%3B%5Cn%7D'%2B'%5Cn'%2B'%5Cnpascal_triangle%285%2C4%29%3B%5Cn%20%20%20%20%20%20%20%20'%2B'%5Cn'%29%3B%20var%20url%20%3D%20'https%3A%2F%2Fsource-academy.github.io%2Fplayground%23chap%3D1%26prgrm%3D'%2Bcompressed%3B%20window.open%28url%29%3B%20%22%3Efunction%20pascal_triangle%28row%2C%20index%29%20%7B%0A%20%20%20%20return%20index%20%3E%20row%0A%20%20%20%20%20%20%20%20%20%20%20%3F%20false%0A%20%20%20%20%20%20%20%20%20%20%20%3A%20index%20%3D%3D%3D%201%20%7C%7C%20index%3D%3D%3Drow%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%3F%201%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%3A%20pascal_triangle%28row%20-%201%2C%20index%20-%201%29%0A%09%20%20%20%20%20%20%20%2B%0A%09%20%20%20%20%20%20%20pascal_triangle%28row%20-%201%2C%20index%29%3B%0A%7D%3C%2Fpre%3E%3C%2Fdiv%3E%3C%2Fdiv%3E%0A%3C%2Fsolution%3E%3C%2Fdiv%3E%0A%3C%2Fdiv%3E%0A%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%3C%2Fexercise%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22ex_1.13%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cexercise%3E%3Cb%3E%3Ca%20class%3D%22exercise-number%20permalink%22%20id%3D%22ex_1.13%22%3EExercise%201.13%20%3C%2Fa%3E%3C%2Fb%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20Prove%20that%20%3Cimg%20alt%3D%22%255Ctextrm%257BFib%257D%2528n%2529%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3D%255Ctextrm%257BFib%257D%2528n%2529%22%20%2F%3E%20is%20the%20closest%20integer%20to%20%3Cimg%20alt%3D%22%255Cphi%255En%252F%255Csqrt%257B5%257D%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3D%255Cphi%255En%252F%255Csqrt%257B5%257D%22%20%2F%3E%2C%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20where%20%3Cimg%20alt%3D%22%255Cphi%253D%2520%25281%252B%255Csqrt%257B5%257D%2529%252F2%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3D%255Cphi%253D%2520%25281%252B%255Csqrt%257B5%257D%2529%252F2%22%20%2F%3E.%20%20Hint%3A%20Let%20%3Cimg%20alt%3D%22%255Cpsi%253D%2520%25281-%255Csqrt%257B5%257D%2529%252F2%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3D%255Cpsi%253D%2520%25281-%255Csqrt%257B5%257D%2529%252F2%22%20%2F%3E.%20%20Use%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20induction%20and%20the%20definition%20of%20the%20Fibonacci%20numbers%20%28see%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20section%C2%A0%3Cref%20name%3D%22sec%3Atree-recursion%22%3E%3Ca%20class%3D%22superscript%22%20id%3D%221.2.2-sec-link-1.2.2%22%20href%3D%2213%22%3E1.2.2%3C%2Fa%3E%3C%2Fref%3E%29%20to%20prove%20that%20](http://latex.codecogs.com/gif.latex?%2C%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20given%20half-dollars%2C%20quarters%2C%20dimes%2C%20nickels%2C%20and%20pennies%3F%20%20More%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20generally%2C%20can%20we%20write%20a%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20function%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20to%20compute%20the%20number%20of%20ways%20to%20change%20any%20given%20amount%20of%20money%3F%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22p8%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cp%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20This%20problem%20has%20a%20simple%20solution%20as%20a%20recursive%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20function.%20%20Suppose%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20we%20think%20of%20the%20types%20of%20coins%20available%20as%20arranged%20in%20some%20order.%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20Then%20the%20following%20relation%20holds%3A%0A%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20The%20number%20of%20ways%20to%20change%20amount%20%3Cimg%20alt%3D%22a%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Da%22%20%2F%3E%20using%20%3Cimg%20alt%3D%22n%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Dn%22%20%2F%3E%20kinds%20of%20coins%20equals%0A%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Cul%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Cli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20the%20number%20of%20ways%20to%20change%20amount%20%3Cimg%20alt%3D%22a%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Da%22%20%2F%3E%20using%20all%20but%20the%20first%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20kind%20of%20coin%2C%20plus%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Fli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Cli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20the%20number%20of%20ways%20to%20change%20amount%20%3Cimg%20alt%3D%22a-d%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Da-d%22%20%2F%3E%20using%20all%20%3Cimg%20alt%3D%22n%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Dn%22%20%2F%3E%20kinds%20of%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20coins%2C%20where%20%3Cimg%20alt%3D%22d%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Dd%22%20%2F%3E%20is%20the%20denomination%20of%20the%20first%20kind%20of%20coin.%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Fli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Ful%3E%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22p9%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cp%3E%0A%20%20%20%20%20%20%20%20%20%20To%20see%20why%20this%20is%20true%2C%20observe%20that%20the%20ways%20to%20make%20change%20can%20be%0A%20%20%20%20%20%20%20%20%20%20divided%20into%20two%20groups%3A%20those%20that%20do%20not%20use%20any%20of%20the%20first%20kind%0A%20%20%20%20%20%20%20%20%20%20of%20coin%2C%20and%20those%20that%20do.%20%20Therefore%2C%20the%20total%20number%20of%20ways%20to%0A%20%20%20%20%20%20%20%20%20%20make%20change%20for%20some%20amount%20is%20equal%20to%20the%20number%20of%20ways%20to%20make%0A%20%20%20%20%20%20%20%20%20%20change%20for%20the%20amount%20without%20using%20any%20of%20the%20first%20kind%20of%20coin%2C%0A%20%20%20%20%20%20%20%20%20%20plus%20the%20number%20of%20ways%20to%20make%20change%20assuming%20that%20we%20do%20use%20the%0A%20%20%20%20%20%20%20%20%20%20first%20kind%20of%20coin.%20%20But%20the%20latter%20number%20is%20equal%20to%20the%20number%20of%0A%20%20%20%20%20%20%20%20%20%20ways%20to%20make%20change%20for%20the%20amount%20that%20remains%20after%20using%20a%20coin%20of%0A%20%20%20%20%20%20%20%20%20%20the%20first%20kind.%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22p10%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cp%3E%0A%20%20%20%20%20%20%20%20%20%20Thus%2C%20we%20can%20recursively%20reduce%20the%20problem%20of%20changing%20a%20given%20amount%0A%20%20%20%20%20%20%20%20%20%20to%20problems%20of%20changing%20smaller%20amounts%20or%20using%20fewer%20kinds%20of%20coins.%0A%20%20%20%20%20%20%20%20%20%20Consider%20this%20reduction%20rule%20carefully%2C%20and%20convince%20yourself%20that%20we%0A%20%20%20%20%20%20%20%20%20%20can%20use%20it%20to%20describe%20an%20algorithm%20if%20we%20specify%20the%20following%0A%20%20%20%20%20%20%20%20%20%20degenerate%20cases%3A%3Ca%20class%3D%22superscript%22%20id%3D%22footnote-link-2%22%20href%3D%2213%23footnote-2%22%3E%5B2%5D%3C%2Fa%3E%0A%0A%20%20%20%20%20%20%20%20%20%20%3Cul%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%3Cli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20If%20%3Cimg%20alt%3D%22a%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Da%22%20%2F%3E%20is%20exactly%200%2C%20we%20should%20count%20that%20as%201%20way%20to%20make%20change.%0A%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Fli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%3Cli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20If%20%3Cimg%20alt%3D%22a%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Da%22%20%2F%3E%20is%20less%20than%200%2C%20we%20should%20count%20that%20as%200%20ways%20to%20make%20change.%0A%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Fli%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%3Cli%3E%20If%20%3Cimg%20alt%3D%22n%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Dn%22%20%2F%3E%20is%200%2C%20we%20should%20count%20that%20as%200%20ways%20to%20make%20change.%0A%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Fli%3E%0A%20%20%20%20%20%20%20%20%20%20%3C%2Ful%3E%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22p11%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cp%3E%0A%20%20%20%20%20%20%20%20%20%20We%20can%20easily%20translate%20this%20description%20into%20a%20recursive%0A%20%20%20%20%20%20%20%20%20%20function%3A%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22snippet%22%20id%3D%22javascript_13_2_div%22%3E%3Cdiv%20class%3D%22pre-prettyprint%22%3E%3Cpre%20class%3D%22prettyprint%22%20title%3D%22Evaluate%20Javascript%20expression%22%20onclick%3D%22var%20compressed%20%3D%20LZString.compressToEncodedURIComponent%28''%2B'%5Cn'%2B'function%20count_change%28amount%29%20%7B%5Cn%20%20%20%20return%20cc%28amount%2C%205%29%3B%5Cn%7D%5Cnfunction%20cc%28amount%2C%20kinds_of_coins%29%20%7B%5Cn%20%20%20%20return%20amount%20%3D%3D%3D%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%3F%201%5Cn%20%20%20%20%20%20%20%20%20%20%20%3A%20amount%20%3C%200%20%7C%7C%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%3F%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%3A%20cc%28amount%2C%20kinds_of_coins%20-%201%29%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%2B%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20cc%28amount%20-%20first_denomination%28%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%29%2C%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%29%3B%5Cn%7D%5Cnfunction%20first_denomination%28kinds_of_coins%29%20%7B%5Cn%20%20%20%20return%20kinds_of_coins%20%3D%3D%3D%201%20%3F%201%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%202%20%3F%205%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%203%20%3F%2010%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%204%20%3F%2025%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%205%20%3F%2050%20%3A%200%3B%5Cn%7D'%2B'%5Cn'%2B'%5Cncount_change%28100%29%3B%5Cn%20%20%20%20%20%20%20%20%20%20'%2B'%5Cn'%29%3B%20var%20url%20%3D%20'https%3A%2F%2Fsource-academy.github.io%2Fplayground%23chap%3D1%26prgrm%3D'%2Bcompressed%3B%20window.open%28url%29%3B%20%22%3Efunction%20count_change%28amount%29%20%7B%0A%20%20%20%20return%20cc%28amount%2C%205%29%3B%0A%7D%0Afunction%20cc%28amount%2C%20kinds_of_coins%29%20%7B%0A%20%20%20%20return%20amount%20%3D%3D%3D%200%0A%20%20%20%20%20%20%20%20%20%20%20%3F%201%0A%20%20%20%20%20%20%20%20%20%20%20%3A%20amount%20%3C%200%20%7C%7C%0A%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%200%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%3F%200%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%3A%20cc%28amount%2C%20kinds_of_coins%20-%201%29%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%2B%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20cc%28amount%20-%20first_denomination%28%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%29%2C%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%29%3B%0A%7D%0Afunction%20first_denomination%28kinds_of_coins%29%20%7B%0A%20%20%20%20return%20kinds_of_coins%20%3D%3D%3D%201%20%3F%201%20%3A%0A%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%202%20%3F%205%20%3A%0A%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%203%20%3F%2010%20%3A%0A%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%204%20%3F%2025%20%3A%0A%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%205%20%3F%2050%20%3A%200%3B%0A%7D%3C%2Fpre%3E%3C%2Fdiv%3E%3C%2Fdiv%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22p12%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cp%3E%0A%20%20%20%20%20%20%20%20%20%20%28The%0A%20%20%20%20%20%20%20%20%20%20%3Ccode%3Efirst_denomination%3C%2Fcode%3E%0A%20%20%20%20%20%20%20%20%20%20function%0A%20%20%20%20%20%20%20%20%20%20takes%20as%20input%20the%20number%20of%0A%20%20%20%20%20%20%20%20%20%20kinds%20of%20coins%20available%20and%20returns%20the%20denomination%20of%20the%20first%0A%20%20%20%20%20%20%20%20%20%20kind.%20%20Here%20we%20are%20thinking%20of%20the%20coins%20as%20arranged%20in%20order%20from%0A%20%20%20%20%20%20%20%20%20%20largest%20to%20smallest%2C%20but%20any%20order%20would%20do%20as%20well.%29%20%20We%20can%20now%0A%20%20%20%20%20%20%20%20%20%20answer%20our%20original%20question%20about%20changing%20a%20dollar%3A%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22snippet%22%20id%3D%22javascript_13_3_div%22%3E%3Cdiv%20class%3D%22pre-prettyprint%22%3E%3Cpre%20class%3D%22prettyprint%22%20title%3D%22Evaluate%20Javascript%20expression%22%20onclick%3D%22var%20compressed%20%3D%20LZString.compressToEncodedURIComponent%28'%5Cnfunction%20count_change%28amount%29%20%7B%5Cn%20%20%20%20return%20cc%28amount%2C%205%29%3B%5Cn%7D%5Cnfunction%20cc%28amount%2C%20kinds_of_coins%29%20%7B%5Cn%20%20%20%20return%20amount%20%3D%3D%3D%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%3F%201%5Cn%20%20%20%20%20%20%20%20%20%20%20%3A%20amount%20%3C%200%20%7C%7C%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%3F%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%3A%20cc%28amount%2C%20kinds_of_coins%20-%201%29%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%2B%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20cc%28amount%20-%20first_denomination%28%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%29%2C%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%29%3B%5Cn%7D%5Cnfunction%20first_denomination%28kinds_of_coins%29%20%7B%5Cn%20%20%20%20return%20kinds_of_coins%20%3D%3D%3D%201%20%3F%201%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%202%20%3F%205%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%203%20%3F%2010%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%204%20%3F%2025%20%3A%5Cn%20%20%20%20%20%20%20%20%20%20%20kinds_of_coins%20%3D%3D%3D%205%20%3F%2050%20%3A%200%3B%5Cn%7D%5Cn%20%20%20%20%20%20%20%20%20%20'%2B'%5Cn'%2B'count_change%28100%29%3B'%2B'%5Cn'%2B''%2B'%5Cn'%29%3B%20var%20url%20%3D%20'https%3A%2F%2Fsource-academy.github.io%2Fplayground%23chap%3D1%26prgrm%3D'%2Bcompressed%3B%20window.open%28url%29%3B%20%22%3Ecount_change%28100%29%3B%3C%2Fpre%3E%3C%2Fdiv%3E%3C%2Fdiv%3E%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22p13%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cp%3E%0A%20%20%20%20%20%20%20%20%20%20The%20function%20%3Ccode%3Ecount_change%3C%2Fcode%3E%0A%20%20%20%20%20%20%20%20%20%20generates%20a%20tree-recursive%20process%20with%0A%20%20%20%20%20%20%20%20%20%20redundancies%20similar%20to%20those%20in%20our%20first%20implementation%20of%0A%20%20%20%20%20%20%20%20%20%20%3Ccode%3Efib%3C%2Fcode%3E.%0A%20%20%20%20%20%20%20%20%20%20%28It%20will%20take%20quite%20a%20while%20for%20that%20292%20to%20be%20computed.%29%20%20On%0A%20%20%20%20%20%20%20%20%20%20the%20other%20hand%2C%20it%20is%20not%20obvious%20how%20to%20design%20a%20better%20algorithm%0A%20%20%20%20%20%20%20%20%20%20for%20computing%20the%20result%2C%20and%20we%20leave%20this%20problem%20as%20a%20challenge.%0A%20%20%20%20%20%20%20%20%20%20The%20observation%20that%20a%0A%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20tree-recursive%20process%20may%20be%20highly%0A%20%20%20%20%20%20%20%20%20%20inefficient%20but%20often%20easy%20to%20specify%20and%20understand%20has%20led%20people%20to%0A%20%20%20%20%20%20%20%20%20%20propose%20that%20one%20could%20get%20the%20best%20of%20both%20worlds%20by%20designing%20a%0A%20%20%20%20%20%20%20%20%20%20%3Cquote%3Esmart%20compiler%3C%2Fquote%3E%20that%20could%20transform%20tree-recursive%0A%20%20%20%20%20%20%20%20%20%20functions%0A%20%20%20%20%20%20%20%20%20%20into%20more%20efficient%0A%20%20%20%20%20%20%20%20%20%20functions%0A%20%20%20%20%20%20%20%20%20%20that%20compute%20the%20same%20result.%3Ca%20class%3D%22superscript%22%20id%3D%22footnote-link-3%22%20href%3D%2213%23footnote-3%22%3E%5B3%5D%3C%2Fa%3E%0A%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%3C%2Fp%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22ex_1.11%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cexercise%3E%3Cb%3E%3Ca%20class%3D%22exercise-number%20permalink%22%20id%3D%22ex_1.11%22%3EExercise%201.11%20%3C%2Fa%3E%3C%2Fb%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20A%20function%20%3Cimg%20alt%3D%22f%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Df%22%20%2F%3E%20is%20defined%20by%20the%20rule%20that%20%3Cimg%20alt%3D%22f%2528n%2529%253Dn%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Df%2528n%2529%253Dn%22%20%2F%3E%20if%20%3Cimg%20alt%3D%22n%253C3%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Dn%253C3%22%20%2F%3E%20and%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Cimg%20alt%3D%22f%2528n%2529%253D%257Bf%2528n-1%2529%257D%252B2f%2528n-2%2529%252B3f%2528n-3%2529%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Df%2528n%2529%253D%257Bf%2528n-1%2529%257D%252B2f%2528n-2%2529%252B3f%2528n-3%2529%22%20%2F%3E%20if%20%3Cimg%20alt%3D%22n%255Cge%25203%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Dn%255Cge%25203%22%20%2F%3E.%20%20Write%20a%20JavaScript%20function%20that%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20computes%20%3Cimg%20alt%3D%22f%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Df%22%20%2F%3E%20by%20means%20of%20a%20recursive%20process.%20%20Write%20a%20function%20that%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20computes%20%3Cimg%20alt%3D%22f%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3Df%22%20%2F%3E%20by%20means%20of%20an%20iterative%20process.%0A%3Cdiv%20class%3D%22Solution%22%3E%0A%3Cdiv%20class%3D%22solution_btn%22%3E%3Cbutton%20class%3D%22btn%20btn-secondary%20solution_btn%22%20href%3D%22%23solution_13_1_div%22%20data-toggle%3D%22collapse%22%3ESolution%3C%2Fbutton%3E%3C%2Fdiv%3E%0A%3Cdiv%20class%3D%22solution_content%20collapse%22%20id%3D%22solution_13_1_div%22%3E%3Csolution%3E%0A%20%20%20%20%20%20%20%3Cdiv%20class%3D%22snippet%22%20id%3D%22javascript_13_4_div%22%3E%3Cdiv%20class%3D%22pre-prettyprint%22%3E%3Cpre%20class%3D%22prettyprint%22%20title%3D%22Evaluate%20Javascript%20expression%22%20onclick%3D%22var%20compressed%20%3D%20LZString.compressToEncodedURIComponent%28''%2B'%5Cn'%2B'%2F%2F%20iterative%20function%5Cnfunction%20f_iterative%28n%29%20%7B%5Cn%20%20%20%20return%20n%20%3C%203%5Cn%20%20%20%20%20%20%20%20%20%20%20%3F%20n%5Cn%20%20%20%20%20%20%20%20%20%20%20%3A%20f_iterative_impl%282%2C%201%2C%200%2C%20n%20-%202%29%3B%5Cn%7D%5Cnfunction%20f_iterative_impl%28a%2C%20b%2C%20c%2C%20count%29%20%7B%5Cn%20%20%20%20return%20count%20%3D%3D%3D%200%5Cn%20%20%20%20%20%20%20%20%20%20%20%3F%20a%5Cn%20%20%20%20%20%20%20%20%20%20%20%3A%20f_iterative_impl%28a%20%2B%202%20*%20b%20%2B%203%20*%20c%2C%20a%2C%20b%2C%20count%20-%201%29%3B%5Cn%7D'%2B'%5Cn'%2B'%5Cnf_iterative%285%29%3B%5Cn%20%20%20%20%20%20%20'%2B'%5Cn'%29%3B%20var%20url%20%3D%20'https%3A%2F%2Fsource-academy.github.io%2Fplayground%23chap%3D1%26prgrm%3D'%2Bcompressed%3B%20window.open%28url%29%3B%20%22%3E%2F%2F%20iterative%20function%0Afunction%20f_iterative%28n%29%20%7B%0A%20%20%20%20return%20n%20%3C%203%0A%20%20%20%20%20%20%20%20%20%20%20%3F%20n%0A%20%20%20%20%20%20%20%20%20%20%20%3A%20f_iterative_impl%282%2C%201%2C%200%2C%20n%20-%202%29%3B%0A%7D%0Afunction%20f_iterative_impl%28a%2C%20b%2C%20c%2C%20count%29%20%7B%0A%20%20%20%20return%20count%20%3D%3D%3D%200%0A%20%20%20%20%20%20%20%20%20%20%20%3F%20a%0A%20%20%20%20%20%20%20%20%20%20%20%3A%20f_iterative_impl%28a%20%2B%202%20*%20b%20%2B%203%20*%20c%2C%20a%2C%20b%2C%20count%20-%201%29%3B%0A%7D%3C%2Fpre%3E%3C%2Fdiv%3E%3C%2Fdiv%3E%0A%0A%3Cdiv%20class%3D%22snippet%22%20id%3D%22javascript_13_5_div%22%3E%3Cdiv%20class%3D%22pre-prettyprint%22%3E%3Cpre%20class%3D%22prettyprint%22%20title%3D%22Evaluate%20Javascript%20expression%22%20onclick%3D%22var%20compressed%20%3D%20LZString.compressToEncodedURIComponent%28''%2B'%5Cn'%2B'%2F%2Frecursive%20function%5Cnfunction%20f_recursive%28n%29%20%7B%5Cn%20%20%20%20return%20n%20%3C%203%5Cn%20%20%20%20%20%20%20%20%20%20%20%3F%20n%5Cn%20%20%20%20%20%20%20%20%20%20%20%3A%20f_recursive%28n%20-%201%29%20%2B%5Cn%09%20%20%20%20%202%20*%20f_recursive%28n%20-%202%29%20%2B%5Cn%09%20%20%20%20%203%20*%20f_recursive%28n%20-%203%29%3B%5Cn%7D'%2B'%5Cn'%2B'%5Cnf_recursive%285%29%3B%5Cn%20%20%20%20%20%20%20'%2B'%5Cn'%29%3B%20var%20url%20%3D%20'https%3A%2F%2Fsource-academy.github.io%2Fplayground%23chap%3D1%26prgrm%3D'%2Bcompressed%3B%20window.open%28url%29%3B%20%22%3E%2F%2Frecursive%20function%0Afunction%20f_recursive%28n%29%20%7B%0A%20%20%20%20return%20n%20%3C%203%0A%20%20%20%20%20%20%20%20%20%20%20%3F%20n%0A%20%20%20%20%20%20%20%20%20%20%20%3A%20f_recursive%28n%20-%201%29%20%2B%0A%09%20%20%20%20%202%20*%20f_recursive%28n%20-%202%29%20%2B%0A%09%20%20%20%20%203%20*%20f_recursive%28n%20-%203%29%3B%0A%7D%3C%2Fpre%3E%3C%2Fdiv%3E%3C%2Fdiv%3E%0A%3C%2Fsolution%3E%3C%2Fdiv%3E%0A%3C%2Fdiv%3E%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%3C%2Fexercise%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22ex_1.12%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cexercise%3E%3Cb%3E%3Ca%20class%3D%22exercise-number%20permalink%22%20id%3D%22ex_1.12%22%3EExercise%201.12%20%3C%2Fa%3E%3C%2Fb%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20The%20following%20pattern%20of%20numbers%20is%20called%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Cem%3EPascal's%20triangle%3C%2Fem%3E.%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Ctreetab%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3Cimg%20alt%3D%22%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%257B%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%255Cbegin%257Barray%257D%257Brrrrrrrrr%257D%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%25201%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2526%25202%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%25263%2520%2520%2526%2520%2520%2520%25263%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%25201%2520%2520%2526%2520%2520%2520%2526%25204%2520%2526%2520%2520%2520%2526%25206%2520%2526%2520%2520%2520%2526%25204%2520%2526%2520%2520%2520%2526%25201%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%255Cldots%2520%2526%2520%2526%2520%2526%2520%2526%2520%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%255Cend%257Barray%257D%257D%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3D%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%257B%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%255Cbegin%257Barray%257D%257Brrrrrrrrr%257D%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%25201%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2526%25202%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%25263%2520%2520%2526%2520%2520%2520%25263%2520%2520%2526%2520%2520%2520%25261%2520%2520%2526%2520%2520%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%25201%2520%2520%2526%2520%2520%2520%2526%25204%2520%2526%2520%2520%2520%2526%25206%2520%2526%2520%2520%2520%2526%25204%2520%2526%2520%2520%2520%2526%25201%2520%2520%2520%255C%255C%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%2520%2520%2526%2520%255Cldots%2520%2526%2520%2526%2520%2526%2520%2526%2520%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%255Cend%257Barray%257D%257D%250A%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%2520%22%20%2F%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%3C%2Ftreetab%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20The%20numbers%20at%20the%20edge%20of%20the%20triangle%20are%20all%201%2C%20and%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20each%20number%20inside%20the%20triangle%20is%20the%20sum%20of%20the%20two%20numbers%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20above%20it.%3Ca%20class%3D%22superscript%22%20id%3D%22footnote-link-4%22%20href%3D%2213%23footnote-4%22%3E%5B4%5D%3C%2Fa%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20Write%20a%20function%20that%20computes%20elements%20of%20Pascal's%20triangle%20by%20means%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20of%20a%20recursive%20process.%0A%3Cdiv%20class%3D%22Solution%22%3E%0A%3Cdiv%20class%3D%22solution_btn%22%3E%3Cbutton%20class%3D%22btn%20btn-secondary%20solution_btn%22%20href%3D%22%23solution_13_2_div%22%20data-toggle%3D%22collapse%22%3ESolution%3C%2Fbutton%3E%3C%2Fdiv%3E%0A%3Cdiv%20class%3D%22solution_content%20collapse%22%20id%3D%22solution_13_2_div%22%3E%3Csolution%3E%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22snippet%22%20id%3D%22javascript_13_6_div%22%3E%3Cdiv%20class%3D%22pre-prettyprint%22%3E%3Cpre%20class%3D%22prettyprint%22%20title%3D%22Evaluate%20Javascript%20expression%22%20onclick%3D%22var%20compressed%20%3D%20LZString.compressToEncodedURIComponent%28''%2B'%5Cn'%2B'function%20pascal_triangle%28row%2C%20index%29%20%7B%5Cn%20%20%20%20return%20index%20%3E%20row%5Cn%20%20%20%20%20%20%20%20%20%20%20%3F%20false%5Cn%20%20%20%20%20%20%20%20%20%20%20%3A%20index%20%3D%3D%3D%201%20%7C%7C%20index%3D%3D%3Drow%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%3F%201%5Cn%20%20%20%20%20%20%20%20%20%20%20%20%20%3A%20pascal_triangle%28row%20-%201%2C%20index%20-%201%29%5Cn%09%20%20%20%20%20%20%20%2B%5Cn%09%20%20%20%20%20%20%20pascal_triangle%28row%20-%201%2C%20index%29%3B%5Cn%7D'%2B'%5Cn'%2B'%5Cnpascal_triangle%285%2C4%29%3B%5Cn%20%20%20%20%20%20%20%20'%2B'%5Cn'%29%3B%20var%20url%20%3D%20'https%3A%2F%2Fsource-academy.github.io%2Fplayground%23chap%3D1%26prgrm%3D'%2Bcompressed%3B%20window.open%28url%29%3B%20%22%3Efunction%20pascal_triangle%28row%2C%20index%29%20%7B%0A%20%20%20%20return%20index%20%3E%20row%0A%20%20%20%20%20%20%20%20%20%20%20%3F%20false%0A%20%20%20%20%20%20%20%20%20%20%20%3A%20index%20%3D%3D%3D%201%20%7C%7C%20index%3D%3D%3Drow%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%3F%201%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%3A%20pascal_triangle%28row%20-%201%2C%20index%20-%201%29%0A%09%20%20%20%20%20%20%20%2B%0A%09%20%20%20%20%20%20%20pascal_triangle%28row%20-%201%2C%20index%29%3B%0A%7D%3C%2Fpre%3E%3C%2Fdiv%3E%3C%2Fdiv%3E%0A%3C%2Fsolution%3E%3C%2Fdiv%3E%0A%3C%2Fdiv%3E%0A%20%20%20%20%20%20%20%20%0A%20%20%20%20%20%20%20%20%3C%2Fexercise%3E%0A%3C%2Fdiv%3E%0A%0A%20%20%20%20%20%20%20%20%3Cdiv%20class%3D%22permalink%22%3E%0A%3Ca%20name%3D%22ex_1.13%22%20class%3D%22permalink%22%3E%3C%2Fa%3E%3Cexercise%3E%3Cb%3E%3Ca%20class%3D%22exercise-number%20permalink%22%20id%3D%22ex_1.13%22%3EExercise%201.13%20%3C%2Fa%3E%3C%2Fb%3E%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20Prove%20that%20%3Cimg%20alt%3D%22%255Ctextrm%257BFib%257D%2528n%2529%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3D%255Ctextrm%257BFib%257D%2528n%2529%22%20%2F%3E%20is%20the%20closest%20integer%20to%20%3Cimg%20alt%3D%22%255Cphi%255En%252F%255Csqrt%257B5%257D%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3D%255Cphi%255En%252F%255Csqrt%257B5%257D%22%20%2F%3E%2C%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20where%20%3Cimg%20alt%3D%22%255Cphi%253D%2520%25281%252B%255Csqrt%257B5%257D%2529%252F2%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3D%255Cphi%253D%2520%25281%252B%255Csqrt%257B5%257D%2529%252F2%22%20%2F%3E.%20%20Hint%3A%20Let%20%3Cimg%20alt%3D%22%255Cpsi%253D%2520%25281-%255Csqrt%257B5%257D%2529%252F2%22%20src%3D%22https%3A%2F%2Fwww.zhihu.com%2Fequation%3Ftex%3D%255Cpsi%253D%2520%25281-%255Csqrt%257B5%257D%2529%252F2%22%20%2F%3E.%20%20Use%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20induction%20and%20the%20definition%20of%20the%20Fibonacci%20numbers%20%28see%0A%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20%20section%C2%A0%3Cref%20name%3D%22sec%3Atree-recursion%22%3E%3Ca%20class%3D%22superscript%22%20id%3D%221.2.2-sec-link-1.2.2%22%20href%3D%2213%22%3E1.2.2%3C%2Fa%3E%3C%2Fref%3E%29%20to%20prove%20that%20) \ textrm {Fib}（n）=（\ phi ^ n-\ psi ^ n）/ \ sqrt {5} $。

<button class="btn btn-secondary solution_btn" data-toggle="collapse" href="#no_solution_13_1_div">Add solution</button>There is currently no solution available for this exercise. This textbook adaptation is a community effort. Do consider contributing by providing a solution for this exercise, using a Pull Request in [Github](https://github.com/source-academy/sicp).

* * *

[[1]](13#footnote-link-1) An example of this was hinted at in section <ref name="sec:evaluating-combinations">[1.1.3](5)</ref>: The interpreter itself evaluates expressions using a tree-recursive process.

[[2]](13#footnote-link-2) For example, work through in detail how the reduction rule applies to the problem of making change for 10 cents using pennies and nickels.

[[3]](13#footnote-link-3) One approach to coping with redundant computations is to arrange matters so that we automatically construct a table of values as they are computed. Each time we are asked to apply the function to some argument, we first look to see if the value is already stored in the table, in which case we avoid performing the redundant computation. This strategy, known as _tabulation_ or _memoization_, can be implemented in a straightforward way. Tabulation can sometimes be used to transform processes that require an exponential number of steps (such as `count-change`) into processes whose space and time requirements grow linearly with the input. See exercise <ref name="ex:memoization">[3.27](60#ex_3.27)</ref>.

[[4]](13#footnote-link-4) The elements of Pascal's triangle are called the _binomial coefficients_, because the ![n](img/493731e423d5db62086d0b8705dda0c8.jpg)th row consists of the coefficients of the terms in the expansion of ![%28x%2By%29%5En](img/26712bd19bf05eb667cc6a9feba7ab25.jpg). This pattern for computing the coefficients appeared in Blaise Pascal's 1653 seminal work on probability theory, _Traité du triangle arithmétique_. According to Knuth (1973), the same pattern appears in the _Szu-yuen Yü-chien_ (<quote>The Precious Mirror of the Four Elements</quote>), published by the Chinese mathematician Chu Shih-chieh in 1303, in the works of the twelfth-century Persian poet and mathematician Omar Khayyam, and in the works of the twelfth-century Hindu mathematician Bháscara Áchárya.

